import pygame

from pygame.constants import *

import data
import marshal
import StringIO

def _pickle_method(method):
         func_name = method.im_func.__name__
         obj = method.im_self
         cls = method.im_class
         return _unpickle_method, (func_name, obj, cls)

def _unpickle_method(func_name, obj, cls):
         for cls in cls.mro():
             try:
                 func = cls.__dict__[func_name]
             except KeyError:
                 pass
             else:
                 break
         return func.__get__(obj, cls)

import copy_reg
import types
copy_reg.pickle(types.MethodType, _pickle_method, _unpickle_method) 

class Engine(object):
    def __init__(self, screen):
        self._screen = screen
        self.running = False
        self.states = []
   
    def getCurrent(self):
        if(len(self.states)):
            return self.states[-1]
        return None
    
    def getScreen(self):
        return self._screen
   
    def mouseEvent(self, event):
        current = self.current
        if current:
            if event.type == MOUSEBUTTONDOWN:
                current.mouseDown(event.pos, event.button)
            elif event.type == MOUSEBUTTONUP:
                current.mouseUp(event.pos, event.button)
            elif event.type == MOUSEMOTION:
                current.mouseMove(event.pos)
   
    def push(self,state):
        state.activate()
        if self.current: self.current.deactivate()
        self.states.append(state)
        
    def pop(self):
        self.current.deactivate()
        self.states.pop()
        if self.current: self.current.activate()
        
    def run(self):
        self.clock = pygame.time.Clock()
        self.clock.tick()
        self.running = True
        
        while(self.running and self.current):
            for event in pygame.event.get():
                if event.type == QUIT:
                    return
                elif event.type == KEYDOWN:
                    if event.key == K_ESCAPE:
                        return
                    self.current.keyEvent(event)
                elif event.type == KEYUP:
                    self.current.keyEvent(event)
                elif (event.type == MOUSEBUTTONDOWN or
                      event.type == MOUSEBUTTONUP or
                      event.type == MOUSEMOTION):
                    self.mouseEvent(event)
            delay = self.clock.tick()
            if self.current: self.current.update(delay)
            
            self.screen.clear()
            if self.current: self.current.paint(self.screen)
            pygame.display.flip()
            
            pygame.time.wait(30)
            
    current = property(getCurrent)
    screen = property(getScreen)

class Screen(object):
    def __init__(self, width=800,height=600, fs=False):
        flags = DOUBLEBUF
        if fs: flags = flags | FULLSCREEN
        self._screen = pygame.display.set_mode((width,height),flags)
        
    def blit(self, surface, loc, source=None):
        if source:
            self._screen.blit(surface,loc, source)
        else:
            self._screen.blit(surface,loc)
        
    def clear(self):
        self._screen.fill( (0,0,0) )
        
    def getCaption(self):
        return pygame.display.get_caption()
        
    def getCenter(self):
        return (self.w / 2, self.h / 2)
        
    def getHeight(self):
        return self._screen.get_height()
        
    def getScreen(self):
        return self._screen
    
    def getSize(self):
        return ( self.w, self.h )
    
    def getWidth(self):
        return self._screen.get_width()
    
    def setCaption(self, caption):
        pygame.display.set_caption(caption)
    
    screen = property(getScreen)
    size = property(getSize)
    caption = property(getCaption, setCaption)
    center = property(getCenter)
    w = property(getWidth)
    h = property(getHeight)
    
class State(object):
    def __init__(self, driver):
        self.driver = driver
        self.all = []
        self.keystates = {}
        
    def __getstate__(self):
        odict = self.__dict__.copy()
        del odict['driver']
        return odict
        
    def activate(self):
        pass    
    
    def add(self, actor):
        self.all.append(actor)
    
    def deactivate(self):
        pass
    
    def keyEvent(self, event):
        for actor in self.all:
            actor.keyEvent(event)
    
    def mouseDown(self, at, button):
        for actor in self.all:
            if actor.rect.collidepoint(at):
                actor.click(button, at)
            
    def mouseMove(self, at):
        for actor in self.all:
            actor.mouseMove(at)
            
    def mouseUp(self, at, button):
        for actor in self.all:
            if actor.rect.collidepoint(at):
                actor.release(button)
    
    def paint(self,screen):
        for actor in self.all:
            actor.paint(screen)
    
    def remove(self, actor):
        try:
            self.all.remove(actor)
        except ValueError:
            pass
    
    def update(self, delay):
        for actor in self.all:
            actor.update(delay)
    
class Actor(object):
    def __init__(self):
        self._rect = pygame.Rect(0,0,0,0)
        
    def click(self,button, at):
        pass
        
    def keyEvent(self, event):
        pass
        
    def getRect(self):
        return self._rect
        
    def mouseMove(self,loc):
        pass
        
    def paint(self,screen):
        pass
    
    def release(self,button):
        pass
    
    def update(self, delay):
        pass
    
    rect = property(getRect)
    
class CompositeActor(Actor):
    def __init__(self):
        Actor.__init__(self)
        self.children = []
        # It'll take a really good monitor to beat this
        self.rect.size = (100000,100000)
    
    def add(self, actor):
        self.children.append(actor)
    
    def click(self,button,at):
        for child in self.children:
            if child.rect.collidepoint(at):
                child.click(button, at)
        
    def keyEvent(self, event):
        for child in self.children:
            child.keyEvent(event)
        
    def mouseMove(self,loc):
        for child in self.children:
            child.mouseMove(loc)
        
    def paint(self,screen):
        for child in self.children:
            child.paint(screen)
    
    def release(self,button):
        for child in self.children:
            child.release(button)
    
    def remove(self, actor):
        try:
            self.children.remove(actor)
        except ValueError:
            pass
    
    def update(self, delay):
        for child in self.children:
            child.update(delay)
    
class Box(CompositeActor):
    def __init__(self, rect):
        CompositeActor.__init__(self)
        self._rect = pygame.Rect(rect)
        self.color = ( 10, 10, 200 )
        self.border = (255, 255,255)
        
    def paint(self, screen):
        screen.screen.fill( self.color, self._rect)
        pygame.draw.rect(screen.screen, self.border, self._rect, 1)
        CompositeActor.paint(self,screen)
    
class Label(Actor):
    def __init__(self,text=" ",color= (255,255,255), size=14,
                 font="freesansbold.ttf"):
        Actor.__init__(self)
        self._color = color
        self.font_name = font
        self.size = size
        self.font = data.load_font(font,size)
        self.text = text
        
    def __getstate__(self):
        odict = self.__dict__.copy()
        del odict['image']
        del odict['font']
        return odict
    
    def __setstate__(self,dict):        
        self.__dict__.update(dict)
        self.font = data.load_font(self.font_name, self.size)
        self.image = self.font.render(self.text, True, self.color)
        
    def getColor(self):
        return self._color
        
    def getText(self):
        return self._text
    
    def paint(self,screen):
        screen.blit(self.image, self.rect.topleft)
    
    def refresh(self):
        self.image = self.font.render(self._text,True, self._color )
        self.rect.size = (self.image.get_width(), self.image.get_height())
    
    def setColor(self, color):
        self._color = color
        self.refresh()
    
    def setText(self, text):
        # It might speed this up to make sure the text is actually new
        self._text = text
        self.refresh()
    
    color = property(getColor, setColor)
    text = property(getText,setText)
    
class Button(Label):
    def __init__(self,text,callback,hotkey=None,color= (255,255,255), size=14,
                 font="freesansbold.ttf"):
        Label.__init__(self,text,color,size,font)
        self.callback = callback
        self.background = None
        self.hotkey = hotkey
            
    def click(self,button, at):
        self.callback(self)
        
    def keyEvent(self, event):
        Label.keyEvent(self, event)
        if event.type == KEYDOWN:
            if event.unicode == self.hotkey:
                self.callback(self)
        
    def paint(self, screen):
        if self.background:
            screen.screen.fill( self.background, self._rect )
        Label.paint(self, screen)